'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = lineChart;
exports.dataWaterline = dataWaterline;

var _d = require('d3');

var d3 = _interopRequireWildcard(_d);

var _moment = require('moment');

var _moment2 = _interopRequireDefault(_moment);

require('moment-timezone');

var _colors = require('@density/ui/variables/colors.json');

var _colors2 = _interopRequireDefault(_colors);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

var ONE_MINUTE_IN_MS = 60 * 1000,
    ONE_HOUR_IN_MS = ONE_MINUTE_IN_MS * 60,
    ONE_DAY_IN_MS = ONE_HOUR_IN_MS * 60;

function lineChart(elem) {
  var props = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var svg = d3.select(elem).append('svg');

  var layers = {
    data: svg.append('g').attr('class', 'layer layer-data'),

    xAxis: svg.append('g').attr('class', 'layer layer-x-axis'),
    yAxis: svg.append('g').attr('class', 'layer layer-y-axis'),

    overlay: svg.append('g').attr('class', 'layer layer-overlay')
  };

  layers.overlay.append('g').attr('class', 'overlay-views');
  layers.overlay.append('g').attr('class', 'overlay-points');

  return function (_ref) {
    var timeZone = _ref.timeZone,
        data = _ref.data,
        overlays = _ref.overlays,
        svgWidth = _ref.svgWidth,
        svgHeight = _ref.svgHeight,
        xAxis = _ref.xAxis,
        yAxis = _ref.yAxis,
        xAxisStart = _ref.xAxisStart,
        xAxisEnd = _ref.xAxisEnd,
        yAxisStart = _ref.yAxisStart,
        yAxisEnd = _ref.yAxisEnd,
        overlayShowPoint = _ref.overlayShowPoint,
        overlayPointRadius = _ref.overlayPointRadius;

    // ------------------------------------------------------------------------
    // SET DEFAULT VALUES
    // ------------------------------------------------------------------------
    timeZone = timeZone || 'UTC';
    data = data.map(function (item, ct) {
      item.name = item.name || 'default';
      item.color = item.color || 'rgba(65, 152, 255, 0.2)';
      item.borderColor = item.borderColor || 'rgb(65, 152, 255)';
      item.verticalBaselineOffset = item.verticalBaselineOffset || 0;

      if (!item.data) {
        throw new Error('Line Chart data item index ' + ct + ' (\'' + item.name + '\') missing a data property. This is required!');
      }

      item.data.forEach(function (j, jct) {
        if (!j.timestamp) {
          throw new Error('Line Chart data item index ' + ct + ' (\'' + item.name + '\') missing a timestamp property within the data property (index ' + jct + '). This is required!');
        }
        if (typeof j.value !== 'number') {
          throw new Error('Line Chart data item index ' + ct + ' (\'' + item.name + '\') missing a numeric value property (actual value: ' + j.value + ') within the data property (index ' + jct + '). This is required!');
        }
      });

      // Verify the order of the data by comparing the first and second values to ensure that order is
      // oldest to newest
      if (item.data.length > 0) {
        if (_moment2.default.utc(item.data[0].timestamp).valueOf() - _moment2.default.utc(item.data[1].timestamp).valueOf() > 0) {
          throw new Error(['Line Chart data item index ' + ct + ' (\'' + item.name + '\') looks to have data in the wrong order', '(item index 0 was newer than item index 1). Please presort data from oldest to newest', 'before passing it to this chart.'].join(' '));
        }
      }

      return item;
    });
    overlays = overlays || [];

    if (typeof svgWidth !== 'number') {
      throw new Error('Line Chart is missing the property \'svgWidth\'. This is required!');
    }
    if (typeof svgHeight !== 'number') {
      throw new Error('Line Chart is missing the property \'svgHeight\'. This is required!');
    }

    /* xAxis */
    /* yAxis */

    /* xAxisStart can be undefined */
    /* xAxisEnd can be undefined */
    /* yAxisStart can be undefined */
    /* yAxisEnd can be undefined */

    overlayShowPoint = typeof overlayShowPoint === 'undefined' ? true : overlayShowPoint;
    overlayPointRadius = typeof overlayPointRadius === 'undefined' ? 4.5 : overlayPointRadius;

    // ------------------------------------------------------------------------
    // ADJUST GRAPH SIZE
    // ------------------------------------------------------------------------

    var topMargin = 76;
    var leftMargin = 56;
    var rightMargin = 20;
    var bottomMargin = 96;

    var graphWidth = svgWidth - leftMargin - rightMargin;
    var graphHeight = svgHeight - topMargin - bottomMargin;

    svg.attr('width', svgWidth);
    svg.attr('height', svgHeight);

    layers.xAxis.attr('transform', 'translate(' + leftMargin + ',' + (topMargin + graphHeight) + ')');
    layers.yAxis.attr('transform', 'translate(' + leftMargin + ',' + topMargin + ')');
    layers.data.attr('transform', 'translate(' + leftMargin + ',' + topMargin + ')');

    // ------------------------------------------------------------------------
    // GATHERING DATA POINTS
    // ------------------------------------------------------------------------
    var dataPoints = {};

    var defaultDataset = data.find(function (i) {
      return i.name === 'default';
    });
    if (defaultDataset.data.length === 0) {
      return;
    }

    dataPoints.firstEvent = defaultDataset.data[0];
    dataPoints.lastEvent = defaultDataset.data.slice(-1)[0];

    dataPoints.firstEventXValue = _moment2.default.utc(dataPoints.firstEvent.timestamp).valueOf(); /* epoch ms */
    dataPoints.firstEventYValue = dataPoints.firstEvent.value;
    dataPoints.lastEventXValue = _moment2.default.utc(dataPoints.lastEvent.timestamp).valueOf(); /* epoch ms */
    dataPoints.lastEventYValue = dataPoints.lastEvent.value;

    var defaultDatasetValues = defaultDataset.data.map(function (i) {
      return i.value;
    });
    dataPoints.eventWithSmallestValue = defaultDataset.data[defaultDatasetValues.indexOf(Math.min.apply(Math, defaultDatasetValues))];
    dataPoints.eventWithLargestValue = defaultDataset.data[defaultDatasetValues.indexOf(Math.max.apply(Math, defaultDatasetValues))];

    dataPoints.startXValue = xAxisStart ? _moment2.default.utc(xAxisStart).valueOf() : dataPoints.firstEventXValue;
    dataPoints.endXValue = xAxisEnd ? _moment2.default.utc(xAxisEnd).valueOf() : dataPoints.lastEventXValue;

    dataPoints.startYValue = typeof yAxisStart !== 'undefined' ? yAxisStart : dataPoints.eventWithSmallestValue.value;
    dataPoints.endYValue = typeof yAxisEnd !== 'undefined' ? yAxisEnd : dataPoints.eventWithLargestValue.value;

    var xScale = d3.scaleLinear().rangeRound([0, graphWidth]).domain([dataPoints.startXValue, dataPoints.endXValue]);

    var yScale = d3.scaleLinear().rangeRound([graphHeight, 0]).domain([dataPoints.startYValue, dataPoints.endYValue]);

    // ------------------------------------------------------------------------
    // X AXIS
    // ------------------------------------------------------------------------
    xAxis({
      scale: xScale,
      timeZone: timeZone,
      bottomMargin: bottomMargin,
      dataPoints: dataPoints
    }, layers.xAxis);

    // ------------------------------------------------------------------------
    // Y AXIS
    // ------------------------------------------------------------------------
    yAxis({
      scale: yScale,
      firstEventYValue: dataPoints.startYValue,
      lastEventYValue: dataPoints.endYValue,
      graphWidth: graphWidth,
      leftMargin: leftMargin,
      dataPoints: dataPoints,
      defaultDataset: defaultDataset
    }, layers.yAxis);

    // ------------------------------------------------------------------------
    // OVERLAY
    // ------------------------------------------------------------------------
    function overlayMouseAction(action, defaultDataset) {
      return function () {
        // Break early if we shouldn't draw the overlay point
        if (!overlayShowPoint) {
          return;
        }

        var selectionData = [],
            x = null,
            y = null;
        if (action === 'mousemove') {
          var coordinates = d3.mouse(this);
          x = coordinates[0] - leftMargin;
          y = coordinates[1] - topMargin;
          selectionData = [x];
        }

        // Render the point on the graph that correspnds to where the user currently has their
        // mouse directed at.
        var pointSelection = layers.overlay.select('.overlay-points').selectAll('.point').data(selectionData);
        pointSelection.enter().append('circle').attr('class', 'point').attr('r', overlayPointRadius).attr('fill', _colors2.default.brandPrimary).merge(pointSelection).attr('cx', function (d) {
          return leftMargin + d;
        }).attr('cy', function (d) {
          var xInMs = xScale.invert(d);
          var eventIndexAtOverlayPosition = d3.bisector(function (d) {
            return _moment2.default.utc(d.timestamp).valueOf();
          }).right(defaultDataset.data, xInMs) - 1;
          var eventAtPosition = defaultDataset.data[eventIndexAtOverlayPosition] || defaultDataset.data[0];

          var renderAfterLastDatapoint = typeof defaultDataset.renderAfterLastDatapoint === 'undefined' ? false : true;
          if (eventAtPosition && (renderAfterLastDatapoint ? true : xInMs <= dataPoints.lastEventXValue)) {
            return topMargin + yScale(eventAtPosition.value) - defaultDataset.verticalBaselineOffset;
          } else {
            return -100000000;
          }
        });
        pointSelection.exit().remove();

        // Render any number of overlays on top of the point.
        var overlaySelection = layers.overlay.select('.overlay-views').attr('transform', 'translate(' + leftMargin + ',' + topMargin + ')').selectAll('.overlay-view').data(
        /* ensure that each overlay exits when the mouse leaves */
        action === 'mouseleave' ? [] : overlays);
        overlaySelection.enter().append('g').attr('class', 'overlay-view').each(function (d) {
          return d.enter({
            /* no mouse x or y here since they are null */
            xScale: xScale,
            yScale: yScale,
            graphWidth: graphWidth,
            graphHeight: graphHeight,
            defaultDataset: defaultDataset,
            dataPoints: dataPoints
          }, d3.select(this)); /* this = reference to the g.overlay-view created above */
        }).merge(overlaySelection).each(function (d) {
          if (x !== null && y !== null) {
            return d.merge({
              mouseX: x,
              mouseY: y,
              xScale: xScale,
              yScale: yScale,
              graphWidth: graphWidth,
              graphHeight: graphHeight,
              defaultDataset: defaultDataset,
              dataPoints: dataPoints
            }, d3.select(this)); /* this = reference to the g.overlay-view created above */
          }
        });
        overlaySelection.exit().each(function (d) {
          return d.exit({
            // no mouse x or y here since they most likely the mouse is outside of the overlay
            // rect at this point
            xScale: xScale,
            yScale: yScale,
            graphWidth: graphWidth,
            graphHeight: graphHeight,
            defaultDataset: defaultDataset,
            dataPoints: dataPoints
          }, d3.select(this)); /* this = reference to the g.overlay-view created above */
        });
      };
    }
    var overlayPointSelection = layers.overlay.selectAll('.overlay').data([dataPoints]);
    overlayPointSelection.enter().append('rect').attr('class', 'overlay').attr('fill', 'transparent').attr('x', leftMargin).attr('y', topMargin).attr('width', graphWidth).attr('height', graphHeight).merge(overlayPointSelection).on('mousemove', overlayMouseAction('mousemove', defaultDataset)).on('mouseleave', overlayMouseAction('mouseleave', defaultDataset));
    overlayPointSelection.exit().remove();

    // ------------------------------------------------------------------------
    // WATERLINES
    // ------------------------------------------------------------------------
    var dataSelection = layers.data.selectAll('.data').data(data);
    dataSelection.enter().append('g').attr('class', 'data').merge(dataSelection).each(function (d) {
      return d.type(d)({
        xScale: xScale,
        yScale: yScale,
        graphHeight: graphHeight,
        graphWidth: graphWidth,
        dataPoints: dataPoints
      }, this); /* this = reference to the g.data created above */
    });
  };
}

function dataWaterline(_ref2) {
  var data = _ref2.data,
      color = _ref2.color,
      borderColor = _ref2.borderColor,
      verticalBaselineOffset = _ref2.verticalBaselineOffset,
      renderAfterLastDatapoint = _ref2.renderAfterLastDatapoint;

  renderAfterLastDatapoint = typeof renderAfterLastDatapoint === 'undefined' ? false : true;

  return function (_ref3, element) {
    var xScale = _ref3.xScale,
        yScale = _ref3.yScale,
        graphHeight = _ref3.graphHeight,
        dataPoints = _ref3.dataPoints;

    var waterlineFillPrefix = 'M0,' + graphHeight + ( /* move to the lower left of the graph */
    'V' + yScale(dataPoints.firstEvent.value)); /* Move to the first datapoint's y value */

    var waterlineStrokePrefix = 'M0,' + yScale(dataPoints.firstEvent.value); /* Move to the first datapoint */

    var waterlineFillPostfix = renderAfterLastDatapoint ? 'H' + xScale(dataPoints.endXValue) + 'V' + graphHeight + 'H0' : 'H' + xScale(dataPoints.lastEventXValue) + 'V' + graphHeight + 'H0';

    var waterlineStrokePostfix = renderAfterLastDatapoint ? 'H' + xScale(dataPoints.endXValue) : 'H' + xScale(dataPoints.lastEventXValue);

    var waterlineSelection = d3.select(element).selectAll('.waterline').data([{
      data: data,
      path: data.reduce(function (acc, i, ct) {
        var timeEpoch = _moment2.default.utc(i.timestamp).valueOf();
        if (timeEpoch > dataPoints.endXValue) {
          return acc;
        }

        var xPosition = xScale(timeEpoch);
        var yPosition = yScale(i.value);
        if (ct === 0) {
          // For the first plotted point, don't draw an outline on the left edge.
          return acc + 'M0,' + (yPosition - verticalBaselineOffset);
        } else if (xPosition > 0) {
          // Plot each point my moving horizontally then vertically.
          return acc + 'H' + xPosition + 'V' + (yPosition - verticalBaselineOffset);
        } else {
          return acc;
        }
      }, '')
    }], function (d) {
      return d.data;
    });
    var waterlineSelectionGroup = waterlineSelection.enter().append('g').attr('class', 'waterline');
    waterlineSelectionGroup.append('path').attr('class', 'waterline-fill').attr('shape-rendering', 'crispEdges').attr('fill', color);
    waterlineSelectionGroup.append('path').attr('class', 'waterline-stroke').attr('stroke', borderColor).attr('shape-rendering', 'crispEdges').attr('fill', 'transparent');

    waterlineSelectionGroup.merge(waterlineSelection).select('.waterline-fill').attr('d', function (d) {
      return waterlineFillPrefix + d.path + waterlineFillPostfix;
    });
    waterlineSelectionGroup.merge(waterlineSelection).select('.waterline-stroke').attr('d', function (d) {
      return waterlineStrokePrefix + d.path + waterlineStrokePostfix;
    });
    waterlineSelection.exit().remove();
  };
}

